from shutil import copyfile
from typing import Literal
from shrinkplz import SHRINKPLZ_DATA
from shrinkplz.config import Config
from shrinkplz.output import perr
from shrinkplz.state import SessionStepState

type MarkResult = Literal["pass"] | Literal["fail"] | Literal["invalid"]


def mark_result_in_log(state: SessionStepState, result: MarkResult):
    with open(SHRINKPLZ_DATA / "session-log", "a") as f:
        f.write(f"{state.bucket_size},{state.cut_idx},{state.drop_count},{result}\n")


def progress_state_if_needed(state: SessionStepState):
    """
    Check to see if we are beyond the end of our data

    if so, we should shrink the bucket size and go to
    the beginning.
    """
    if state.cut_idx >= state.current_smallest:
        state.cut_idx = 0
        state.bucket_size //= 2
        state.drop_count = 0


def mark_input(state: SessionStepState, result: MarkResult) -> bool:
    """
    Mark some input

    returns whether we are "done"
    """
    mark_result_in_log(state, result)
    match result:
        case "pass":
            # if we passed that means the cut data was important
            # so we're going to re-insert it, then move the cut_idx forward
            state.cut_idx += state.bucket_size
        case "fail":
            # if we fail, that means we're in a good spot
            # we can discard it and increment the drop_count
            state.drop_count += 1
            new_linecount = commit_pending_drop(state.cut_idx, state.bucket_size)
            state.current_smallest = new_linecount
        case "invalid":
            # if we have invalid data we can't say whether we're in
            # a good state or not, so we'll just move the cut_idx forward
            state.cut_idx += state.bucket_size

    progress_state_if_needed(state)
    if state.looks_completed():
        copyfile(SHRINKPLZ_DATA / "current-smallest", "current-input")
        perr(
            """
Completed our search!
The smallest known data is copied
to current-input
"""
        )
        return True
    save_state(state)
    # let's prepare the current-input
    cut_input_data(state.cut_idx, state.bucket_size)
    perr("Smallest failed test data written to current-input")
    return False


def commit_pending_drop(cut_idx: int, bucket_size: int) -> int:
    """
    Commit the pending drop on the current-smallest file
    """
    # We redo the logic to generate the current-input instead of copying it
    # to avoid issues with current-input being written to
    with open(SHRINKPLZ_DATA / "current-smallest", "r") as f:
        current_smallest = f.readlines()
    shrunk = current_smallest[:cut_idx] + current_smallest[cut_idx + bucket_size :]
    new_linecount = len(shrunk)
    with open(SHRINKPLZ_DATA / "current-smallest", "w") as f:
        f.writelines(shrunk)

    return new_linecount


def save_state(state: SessionStepState):
    with open(SHRINKPLZ_DATA / "current-state", "w") as f:
        state.write_into_file(f)


def read_state() -> SessionStepState:
    with open(SHRINKPLZ_DATA / "current-state", "r") as f:
        return SessionStepState.read_from_file(f)


def cut_input_data(cut_idx: int, bucket_size: int):
    """
    Given our state, cut the input data, holding onto the cut data
    in a pending-drop file

    reads current-smallest, generates pending-drop and current-input
    """
    # this is not very smart
    with open(SHRINKPLZ_DATA / "current-smallest", "r") as f:
        current_smallest = f.readlines()
    # split out the data

    # first, we have our pending drop
    pending_drop = current_smallest[cut_idx : cut_idx + bucket_size]
    with open(SHRINKPLZ_DATA / "pending-drop", "w") as f:
        f.writelines(pending_drop)

    # then we build out our current input
    current_input = (
        current_smallest[:cut_idx] + current_smallest[cut_idx + bucket_size :]
    )

    # XXX current-input target
    with open("current-input", "w") as f:
        f.writelines(current_input)



import subprocess
import tempfile
import pytest
from pathlib import Path
import shutil
import os

ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

proj_root = Path(__file__).parent.parent


@pytest.fixture
def temp_dir():
    """
    Fixture that provides a TempDirectory instance with automatic cleanup.
    """
    temp_dir = Path(tempfile.mkdtemp())
    yield temp_dir
    shutil.rmtree(temp_dir)


@pytest.fixture
def run_shrinkplz():
    """
    Fixture that provides a helper functon to run shrinkplz commands.
    """

    def _run_command(args, cwd=None, input_data=None):
        """
        Run shrinkplz command with given arguments.

        Args:
            args: List of command line arguments (without 'shrinkplz')
            cwd: Working directory to run command in
            input_data: Data to pass to stdin

        Returns:
            subprocess.CompletedProcess result
        """
        cmd = ["uv", "run", "shrinkplz"] + args
        result = subprocess.run(
            cmd, cwd=cwd, capture_output=True, text=True, input=input_data
        )
        return result

    return _run_command


class ScriptSession:
    """
    A shrinkplz script session
    """

    def __init__(self, temp_dir: Path) -> None:
        self.temp_dir = temp_dir

        # copy over the test prog and the test script
        shutil.copy(proj_root / "example" / "my-prog.py", self.temp_dir / "my-prog.py")
        shutil.copy(
            proj_root / "example" / "my-test-script", self.temp_dir / "my-test-script"
        )

    def set_failure_cases(self, cases: str) -> None:
        """
        Set the failure cases for this session
        (one char turns into one line)
        """
        with open(self.temp_dir / "failure-cases", "w") as f:
            f.writelines([f"{c}\n" for c in cases])

    def set_starting_input(self, start: str) -> None:
        """
        Set our starting input
        (one char turns into one line)
        """
        with open(self.temp_dir / "starting-test-input", "w") as f:
            f.writelines([f"{c}\n" for c in start])

    def run(
        self, args: list[str] | None = None
    ) -> tuple[subprocess.CompletedProcess, list[str]]:
        # TODO I should just call the command directly...

        if args is None:
            args = []
        cmd = (
            ["uv", "run", "shrinkplz"]
            + args
            + ["script", "./my-test-script", "./starting-test-input"]
        )
        result = subprocess.run(
            cmd,
            cwd=self.temp_dir,
            text=True,
        )

        assert result.returncode == 0

        final_lines = [
            line.strip()
            for line in open(self.temp_dir / "current-input", "r").readlines()
        ]

        return final_lines


@pytest.fixture
def session(temp_dir):
    yield ScriptSession(temp_dir)


def test_shrinkplz_script_session(session: ScriptSession):
    """
    Test shrinkplz script session using the example data.
    """

    session.set_failure_cases("DUVX")
    session.set_starting_input(ALPHABET)

    final_lines = session.run()

    assert final_lines == ["D", "U", "V", "X"]


def test_minimum_size_session(session: ScriptSession):
    session.set_failure_cases("DUVX")
    session.set_starting_input(ALPHABET)

    final_lines = session.run(["--min-test-size", "6"])

    assert len(final_lines) == 6
    # but we should have our failures!
    assert {"D", "U", "V", "X"}.issubset(set(final_lines))



from dataclasses import dataclass


@dataclass
class Config:
    """
    Config state of the shrinkplz run
    """

    # The minimimum size of a test, we'll consider shrinking done at this point
    min_test_size: int



from dataclasses import dataclass
import io

from shrinkplz.config import Config


@dataclass
class SessionStepState:
    """
    This is the state of the current session
    """

    # HACK for now hold onto the config here
    # at some later state this might be problematic
    # and we should then move it off
    config: Config

    # the size of an individual bucket
    bucket_size: int
    # the current index in the file where we should cut
    # next (or where we should re-insert cut data)
    cut_idx: int
    # the number of drops we've done at this bucket size
    drop_count: int
    # the size of our current-smallest data
    current_smallest: int

    @staticmethod
    def read_from_file(config, f: io.FileIO) -> "SessionStepState":
        bucket_size = int(f.readline())
        cut_idx = int(f.readline())
        drop_count = int(f.readline())
        current_smallest = int(f.readline())
        return SessionStepState(
            config=config,
            bucket_size=bucket_size,
            cut_idx=cut_idx,
            drop_count=drop_count,
            current_smallest=current_smallest,
        )

    def write_into_file(self, f: io.FileIO) -> None:
        for line in [
            str(self.bucket_size),
            str(self.cut_idx),
            str(self.drop_count),
            str(self.current_smallest),
        ]:
            f.write(f"{line}\n")

    def looks_completed(self) -> bool:
        return (
            self.bucket_size == 0 or self.config.min_test_size >= self.current_smallest
        )



import shutil
import subprocess
from shrinkplz.config import Config
from shrinkplz.output import perr
from shrinkplz.state import SessionStepState
from shrinkplz.mark import cut_input_data, mark_input, MarkResult
import argparse

import os
from pathlib import Path

parser = argparse.ArgumentParser(
    description="""
Bisect your test data to find the smallest size.

During a shrinking session, mark data as passing or
failing. Shrinkplz will shrink down your data as
much as possible until it finds the smallest data
that still causes a test failure.
"""
)

parser.add_argument(
    "--min-test-size",
    type=int,
    default=1,
    help="""
The minimum size of a test (in number of lines).

defaults to 1.

We won't shrink below the number of lines passed in this argument.
""",
)
subparsers = parser.add_subparsers(dest="cmd_name")

start_args = subparsers.add_parser("start", help="Start a shrinking session")
start_args.add_argument("file_path")

subparsers.add_parser("abandon", help="Abandon a shrinking session")
mark_args = subparsers.add_parser("mark", help="Mark the current input")
mark_args.add_argument("result", choices=["pass", "fail", "invalid"])

script_args = subparsers.add_parser("script", help="Run a scripted shrinking session")
script_args.add_argument("script_path")
script_args.add_argument("file_path")


SHRINKPLZ_DATA = Path(".shrinkplz")


def save_state(state: SessionStepState):
    with open(SHRINKPLZ_DATA / "current-state", "w") as f:
        state.write_into_file(f)


def read_state(config: Config) -> SessionStepState:
    with open(SHRINKPLZ_DATA / "current-state", "r") as f:
        return SessionStepState.read_from_file(config, f)


def mark_cmd(config: Config, result: MarkResult) -> bool:
    """
    Mark command, returns whether we are 'done'
    """
    try:
        state = read_state(config)
    except FileNotFoundError:
        perr("Failed to read the state, are we actually in a session?")
        return True
    return mark_input(state, result)


def start_cmd(config, file_path: str):
    if SHRINKPLZ_DATA.exists():
        perr("A session already exists! Use abandon to get rid of it")
        return 1
    perr("Starting session...")
    with open(file_path, "r") as f:
        initial_data = f.readlines()
    os.mkdir(SHRINKPLZ_DATA)
    with open(SHRINKPLZ_DATA / "current-smallest", "w") as f:
        f.writelines(initial_data)

    initial_line_count = len(initial_data)
    state = SessionStepState(
        config=config,
        bucket_size=initial_line_count // 2,
        cut_idx=0,
        drop_count=0,
        current_smallest=initial_line_count,
    )
    cut_input_data(state.cut_idx, state.bucket_size)
    perr("Wrote first test data to current-input")
    save_state(state)


def script_cmd(config, file_path, script_path):
    """
    Run everything automatically
    """
    start_cmd(config, file_path)
    # now to run everything in a loop

    done = False
    while not done:
        result = subprocess.run(script_path)
        if result.returncode == 0:
            done = mark_cmd(config, "pass")
        elif result.returncode == 125:
            done = mark_cmd(config, "invalid")
        else:
            done = mark_cmd(config, "fail")


def main():
    args = parser.parse_args()
    config = Config(min_test_size=args.min_test_size)
    match cmd_name := args.cmd_name:
        case "mark":
            mark_cmd(config, args.result)
        case "start":
            start_cmd(config, args.file_path)
        case "abandon":
            shutil.rmtree(SHRINKPLZ_DATA)
            perr("abandoned run")
        case "script":
            script_cmd(config, args.file_path, args.script_path)
        case None:
            parser.print_help()
        case _:
            # in theory we shouldn't hit this
            raise AssertionError(f"BUG: Unknown command {cmd_name}")
